from __future__ import unicode_literals
from collections import OrderedDict
import types

import pqaut.automator.factory as factory
import pqaut.automator.qobject
import pqaut.server

class QWidgetAutomator(pqaut.automator.qobject.QObjectAutomator):

    def __init__(self, target):
        self._target = target

    def click(self):
        pqaut.server.clicker.click_on(self._target)

    def clickable_target(self, point):
        return self.target.childAt(point.x(), point.y())

    def global_position(self):
        return self.target.mapToGlobal(self.target.pos())

    def is_offscreen(self):
        return self.target.visibleRegion().isEmpty()

    def is_match(self, value=None, automation_id=None, automation_type=None):
        text = self.get_value()
        nbsp = "\u00A0"
        text_with_normalized_spaces = text.replace(nbsp, " ")
        print('matching value: {}, automation_id: {}, automation_type: {}'.format(value, automation_id, automation_type))

        if automation_type is not None and len(automation_type) > 0 and automation_type != self.automation_type():
            return False

        if value  in [text_with_normalized_spaces, text, self.get_name()]:
                return True

        if automation_id is not None and automation_id == self.automation_id():
                return True

        return False

    def to_json(self, is_recursive=False):
        json=OrderedDict([
                ('type',self.target.__class__.__name__), 
                ('automation_id', self.automation_id()),
                ('automation_type',self.automation_type()),
                ('name',self.get_name()), 
                ('value',self.get_value()), 
                ('visible',self.value_or_default('isVisible', True)),
                ('enabled',self.value_or_default('isEnabled', False)), 
                ('offscreen',self.is_offscreen()),
                ('frame', OrderedDict([
                    ('x',self.value_or_default('x', 0)),
                    ('y',self.value_or_default('y', 0)),
                    ('width',self.value_or_default('width', 0)),
                    ('height',self.value_or_default('height', 0)),
                ])), 
                ('global_position', OrderedDict([
                    ('x',self.global_position().x()),
                    ('y',self.global_position().y()),
                ])),
            ])

        if is_recursive:
            children_json = []
            for child in self.get_children():
                children_json.append(child.to_json(is_recursive))
            json['children'] = children_json

        return json




from __future__ import unicode_literals
import os
import bottle
import json
import threading
import logging
import time

from bottle import template
from PyQt5.Qt import QApplication, QQuickWindow, QTest
from PyQt5.QtCore import Qt

import pqaut.automator.factory as factory
import pqaut.clicker
import pqaut.key_press


@bottle.get("/ping")
def ping():
    if get_root_widget() is None:
        bottle.abort(503, "Still booting up, try again later")

    return 'Ping!'

@bottle.post("/click")
def click():
    return_value = {}
    window_name = get_query_value('window_name')
    value = get_query_value('value')
    automation_type = get_query_value('automation_type')

    try:
        widget = find_widget_in(get_root_widget(window_name), value=value, automation_id=value, automation_type=automation_type)
        if widget is not None:
            return_value = widget.to_json()
            widget.click()
    except Exception as error:
        logger.error("/click got error {0}".format(error))

    return return_value

@bottle.post("/input")
def input():
    found_widget = {}
    window_name = get_query_value('window_name')
    input_item = get_query_value('input_item')
    value = get_query_value('value')
    automation_type = get_query_value('automation_type')
    widget = find_widget_in(get_root_widget(window_name), value=input_item, automation_id=input_item, automation_type=automation_type)

    if widget is not None:
        widget.set_value(value)
        found_widget = widget.to_json()

    return found_widget

@bottle.post("/set_value")
def set_value():
    found_widget = {}
    window_name = get_query_value('window_name')
    automation_id = get_query_value('automation_id')
    property = get_query_value('property')
    value = get_query_value('value')

    widget = find_widget_in(get_root_widget(window_name), automation_id=automation_id)

    if widget is not None:
        widget.set_value(value, property)
        found_widget = widget.to_json()

    return found_widget

@bottle.post("/find_element")
def find_element():
    found_json = {}
    window_name = get_query_value('window_name')
    value = get_query_value('value')
    automation_type = get_query_value('automation_type')

    widget = find_widget_in(get_root_widget(window_name), value=value, automation_id=value, automation_type=automation_type)
    if widget is not None:
        found_json = widget.to_json()

    logger.debug('found element: {}'.format(found_json))
    return found_json

@bottle.get("/")
def ui_tree():
    return template("<pre>{{ui_tree}}</pre>", ui_tree=json.dumps(get_root_widget().to_json(is_recursive=True), sort_keys=False, indent=4, separators=(',',': ')))

def get_query_value(name):
    if bottle.request.json is not None:
        return bottle.request.json.get('query', {}).get(name, '')

    return ''

def get_root_widget(window_name = ''):
    root_widget = None

    quick_windows = [window for window in QApplication.topLevelWindows() if isinstance(window, QQuickWindow)]
    logger.debug('root QQuickWindows: {}, root widgets: {}'.format(quick_windows, QApplication.topLevelWidgets()))
    for root_widget in quick_windows:
        root_widget = factory.automate(root_widget)
        if len(window_name) > 0 and root_widget.is_match(window_name):
            break

    if root_widget is None and any(quick_windows):
        root_widget = factory.automate(quick_windows[0])

    if root_widget is None:
        for root_widget in QApplication.topLevelWidgets():
            root_widget = factory.automate(root_widget)
            if len(window_name) > 0 and root_widget.is_match(window_name):
                break

    if root_widget is None and any(QApplication.topLevelWidgets()):
        root_widget = factory.automate(QApplication.topLevelWidgets()[0])

    logger.debug('found root widget: {}'.format(root_widget if root_widget is None else root_widget.target))
    return root_widget

def find_widget_in(parent, value=None, automation_id=None, automation_type=None):
    if parent is None:
        return None

    for child in parent.get_children():
        if child.is_match(value=value, automation_id=automation_id, automation_type=automation_type):
            return child

        found_widget = find_widget_in(child, value=value, automation_id=automation_id, automation_type=automation_type)
        if found_widget is not None:
            return found_widget

    return None

def start_automation_server():
    #debug = log_level_name is 'DEBUG'
    debug = True
    if debug:
        logger.info("Starting bottle in debug mode")
    thread = threading.Thread(target=bottle.run, kwargs={'host':'0.0.0.0', 'port':5123, 'quiet':(not debug), 'debug':debug})
    thread.setDaemon(True)
    thread.start()


log_levels = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
}
log_level_name = os.environ.get('PQAUT_LOG', 'ERROR')
logging.basicConfig(level=log_levels[log_level_name])
logger = logging.getLogger(__name__)
logger.setLevel(log_levels[log_level_name])
logger.info('pqaut log level is {}'.format(log_level_name))
clicker = pqaut.clicker.Clicker()
key_press = pqaut.key_press.KeyPress()



from __future__ import unicode_literals
import os
import logging

from PyQt5.Qt import QObject, QWidget, QPoint, QPointF, QTest, Qt, pyqtSignal, QApplication, QVariant, pyqtSlot


logger = logging.getLogger(__name__)

class PropertySetter(QObject):
    def __init__(self, target, property, value):
        super(QObject, self).__init__()
        self.target = target
        self.property = property
        self.value = value

    def set_value(self):
        raise NotImplementedError()

class KeyPress(QObject):
    set_value = pyqtSignal(QObject)

    def __init__(self):
        super(QObject, self).__init__()
        self.set_value.connect(self._set_value_on_ui_thread, type=Qt.QueuedConnection)

    def input(self, property_setter):
        logger.debug("setting property {} to {} ".format(property_setter.property, property_setter.value))
        self.set_value.emit(property_setter)

    def _set_value_on_ui_thread(self, property_setter):
        try:
            property_setter.set_value()
        except Exception as ex:
            logger.debug(ex)




from __future__ import unicode_literals
from collections import OrderedDict
import sip
import logging

from PyQt5.Qt import QPointF, QPoint, QObject,  QVariant
import PyQt5.Qt as Qt

import pqaut.automator.factory as factory
import pqaut.automator.qobject
import pqaut.server
import pqaut.key_press

logger = logging.getLogger(__name__)

class QQuickPropertySetter(pqaut.key_press.PropertySetter):

    def set_value(self):
        try:
            meta_object = self.target.metaObject()

            index = meta_object.indexOfProperty(self.property)
            if index < 0:
                raise Exception("could not set property {}".format(self.property))

            meta_object.property(index).write(self.target, QVariant(self.value))
        except Exception as ex:
            logger.error("could not enter text into input item: {}".format(ex))



class QQuickItemAutomator(pqaut.automator.qobject.QObjectAutomator):

    def __init__(self, target):
        self._target = target

    def click(self):
        pointf = self.target.mapToScene(QPointF(0.0, 0.0))
        x = pointf.x()
        y = pointf.y()
        x += self.value_or_default("width", 0.0) / 2.0
        y += self.value_or_default("height", 0.0) / 2.0
        point = QPoint(x,y)
        root_widget = pqaut.server.get_root_widget()
        pqaut.server.clicker.click_on(root_widget.target, point)

    def global_position(self):
        return self.target.mapToScene(QPointF(0.0, 0.0))

    def is_offscreen(self):
        first_child = pqaut.server.get_root_widget()
        root_width = first_child.value_or_default('width', 0.0)
        root_height = first_child.value_or_default('height', 0.0)
        pointf = self._target.mapToScene(QPointF(0.0, 0.0))
        x = pointf.x()
        y = pointf.y()
        width = self.value_or_default("width", 0.0)
        height = self.value_or_default("height", 0.0)
        return x+width <= 0 or x >= root_width or y+height <= 0 or y >= root_height

    def get_children(self):
        children = []

        try:
            children = self._target.childItems()
            if children is None:
                children = self._target.findChildren(QObject)
        except Exception as e:
            pass

        return [factory.automate(c) for c in children];

    def is_match(self, value=None, automation_id=None, automation_type=None):
        text = self.get_value()
        nbsp = "\u00A0"
        text_with_normalized_spaces = text.replace(nbsp, " ")
        assert automation_id is not None

        if automation_type is not None and len(automation_type) > 0 and automation_type != self.automation_type():
            return False

        if value  in [text_with_normalized_spaces, text, self.get_name()]:
            return True

        if automation_id is not None and automation_id == self.automation_id():
            return True

        return False

    def to_json(self, is_recursive = False):
        json=OrderedDict([
                ('type',self._target.__class__.__name__),
                ('automation_id', self.automation_id()),
                ('automation_type',self.automation_type()),
                ('name',self.get_name()),
                ('value',self.get_value()),
                ('visible',self.value_or_default('visible', True)),
                ('enabled',self.value_or_default('enabled', False)),
                ('offscreen',self.is_offscreen()),
                ('frame', OrderedDict([
                    ('x',self.value_or_default('x', 0)),
                    ('y',self.value_or_default('y', 0)),
                    ('width',self.value_or_default('width', 0)),
                    ('height',self.value_or_default('height', 0)),
                ])),
                ('global_position', OrderedDict([
                    ('x',self.global_position().x()),
                    ('y',self.global_position().y()),
                ])),
            ])

        if is_recursive:
            children_json = []
            for child in self.get_children():
                children_json.append(child.to_json(is_recursive))
            json['children'] = children_json

        return json

    def value_or_default(self, value_name, default):
        value = default

        try:
            meta_object = self._target.metaObject()

            index = meta_object.indexOfProperty(value_name)
            if index >= 0:
                value = meta_object.property(index).read(self._target)

            return value
        except Exception as ex:
            return default

    def set_value(self, value, property='text'):
        property_setter = QQuickPropertySetter(self._target, property, value)
        sip.transferto(property_setter, self._target)
        try:
            pqaut.server.key_press.input(property_setter)
        except Exception as ex:
            logger.debug("error on  set value on ui thread: {}".format(ex))

